
// WrapArmsPad3D.java
// Andrew Davison, October 2006, ad@fivedots.coe.psu.ac.th

/* A world full of 3D models and ground shapes on a checkboard floor. 
   The models and ground shapes were first seen in the ObjView3D
   example. They use the ModelLoader and GroundShape classes.

   A new feature is that the models/shapes may include
   translucent boxes, acting as bounding boxes, which are stored
   as Obstacle objects inside Obstacles.

   The user's viewpoint is bracketed by two grabbers (first seen
   in the Arms3D example), which can be moved using the keyboard
   or game pad. The Grabbers object is connected to the viewpoint
   as a PlatformGeometry in addGrabbers().

   The keyboard is managed by the KeyBehavior object, the game pad
   by GamePadController (first seen in the GamePadViewer example)
   and GamePadBehavior.

   If the grabbers hit an obstacle, a sound is played using the
   SoundPlayer object. It should also be playing MIDI music in 
   the background, but a bug in some versions of Java SE 5/6 may 
   prevent it from being heard.
*/


import javax.swing.*;
import java.awt.*;

import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.image.*;
import javax.media.j3d.*;
import javax.vecmath.*;



public class WrapArmsPad3D extends JPanel
// Holds the 3D canvas
{
  private static final int PWIDTH = 512;   // size of panel
  private static final int PHEIGHT = 512; 

  private static final int BOUNDSIZE = 100;  // larger than world

  // info used to position initial viewpoint
  private final static double Z_START = 9.0;

  private static final String BG_DIR = "images/";  
               // directory location of background texture

  private static final String BG_MUSIC = "Mission_Impossible_TV.mid";


  private SimpleUniverse su;
  private BranchGroup sceneBG;
  private BoundingSphere bounds;   // for environment nodes

  private Obstacles obstacles;
  private SoundsPlayer soundsPlayer;


  public WrapArmsPad3D()
  // A panel holding a 3D canvas
  {
    setLayout( new BorderLayout() );
    setOpaque( false );
    setPreferredSize( new Dimension(PWIDTH, PHEIGHT));

    GraphicsConfiguration config =
					SimpleUniverse.getPreferredConfiguration();
    Canvas3D canvas3D = new Canvas3D(config);
    add("Center", canvas3D);
    canvas3D.setFocusable(true);     // give focus to the canvas 
    canvas3D.requestFocus();

    soundsPlayer = new SoundsPlayer();
    soundsPlayer.load(BG_MUSIC);   // load the background MIDI sequence
    soundsPlayer.playLoop(BG_MUSIC);  // play it continuously
    // soundsPlayer.play("mcdonald.mid");
    // soundsPlayer.playLoop("door.wav");

    su = new SimpleUniverse(canvas3D);

    // depth-sort transparent objects on a per-geometry basis
    View view = su.getViewer().getView();
    view.setTransparencySortingPolicy(View.TRANSPARENCY_SORT_GEOMETRY);

    createSceneGraph();
    su.addBranchGraph( sceneBG );
  } // end of WrapArmsPad3D()



  private void createSceneGraph()
  // initilise the scene
  { 
    sceneBG = new BranchGroup();
    bounds = new BoundingSphere(new Point3d(0,0,0), BOUNDSIZE);  

    lightScene();         // add the lights
    addBackground("lava.jpg");
    sceneBG.addChild( new CheckerFloor().getBG() );  // add the floor

    obstacles = new Obstacles(sceneBG, soundsPlayer);
    addModels();
    addGroundCover();

    positionViewPoint();
    addGrabbers();

    sceneBG.compile();   // fix the scene
  } // end of createSceneGraph()


  private void lightScene()
  /* One ambient light, 2 directional lights */
  {
    Color3f white = new Color3f(1.0f, 1.0f, 1.0f);

    // Set up the ambient light
    AmbientLight ambientLightNode = new AmbientLight(white);
    ambientLightNode.setInfluencingBounds(bounds);
    sceneBG.addChild(ambientLightNode);

    // Set up the directional lights
    Vector3f light1Direction  = new Vector3f(-1.0f, -1.0f, -1.0f);
       // left, down, backwards 
    Vector3f light2Direction  = new Vector3f(1.0f, -1.0f, 1.0f);
       // right, down, forwards

    DirectionalLight light1 = 
            new DirectionalLight(white, light1Direction);
    light1.setInfluencingBounds(bounds);
    sceneBG.addChild(light1);

    DirectionalLight light2 = 
        new DirectionalLight(white, light2Direction);
    light2.setInfluencingBounds(bounds);
    sceneBG.addChild(light2);
  }  // end of lightScene()



  private void addModels()
  /* The ModelLoader getModel() method expects a OBJ filename, for
     files in the Models/ subdirectory, and an optional y-axis 
     translation. The method returns a TransformGroup which can
     be used with other TGs to further position, rotate, or scale
     the model.

     Also create an obstacle for each model, and load a sound used
     when the grabbers collide with the obstacle.
  */
  {
    ModelLoader ml = new ModelLoader();
    Transform3D t3d = new Transform3D();

    // a cow model
    t3d.setTranslation( new Vector3d(-2,0,-4));   // move
    t3d.setScale(0.7); // shrink
    TransformGroup tg1 = new TransformGroup(t3d);
    tg1.addChild( ml.getModel("cow.obj", 1.3) );
    sceneBG.addChild(tg1);

    soundsPlayer.load("cow.wav", 1500);
    obstacles.add("cow.wav", -1.75,-4, 2.55,1.6,1);  
       // the args: obstacle sound fnm, (x,z) position, (x,y,z) side lengths

    // a teapot model
    t3d.set( new Vector3d(5,0,-5)); 
    t3d.setScale(0.5);
    TransformGroup tg2 = new TransformGroup(t3d);
    tg2.addChild( ml.getModel("teapot.obj") );
    sceneBG.addChild(tg2);

    soundsPlayer.load("wood.wav", 470);
    obstacles.add("wood.wav", 5,-5, 1,1,1);

    // a pig model
    t3d.set( new Vector3d(4,0,0)); 
    t3d.setScale(0.5); 
    t3d.setRotation( new AxisAngle4d(0,1,0, Math.toRadians(180)) ); 
              // rotate 180 degrees anticlockwise around y-axis
    TransformGroup tg3 = new TransformGroup(t3d);
    tg3.addChild( ml.getModel("pig.obj", 0.8) );
    sceneBG.addChild(tg3);

    soundsPlayer.load("oink.wav", 740);
    obstacles.add("oink.wav", 4,0, 2,1.2,0.5);
  }  // end of addModels()


  private void addGroundCover()
  /* GroundShape expects a GIF filename, located in the images/
     subdirectory and a scale factor. The resulting shape is
     located at (0,0) on the (x, z) plain, and can be adjusted
     with an additional TG.

     As in addModels(), also create an obstacle for each shape, 
     and load a sound used when the grabbers collide with 
     the obstacle.
  */
  { Transform3D t3d = new Transform3D();

    t3d.set( new Vector3d(-3,0,0));
    TransformGroup tg2 = new TransformGroup(t3d);
    tg2.addChild( new GroundShape("tree2.gif", 2) );
    sceneBG.addChild(tg2);

    // the sound when hitting a tree is like hitting a door!
    soundsPlayer.load("door.wav", 480);
    obstacles.add("door.wav", -3,0, 0.3,2,0.3); 


    t3d.set( new Vector3d(2,0,-6));
    TransformGroup tg3 = new TransformGroup(t3d);
    tg3.addChild( new GroundShape("tree4.gif", 3) );
    sceneBG.addChild(tg3);

    obstacles.add("door.wav", 2,-6, 0.3,3,0.3);   
       // no need to load the "door" sound again
  }  // end of addGroundCover()


  private void positionViewPoint()
  // position the camera viewpoint
  {
    // get targetTG, the TG for the viewpoint
    ViewingPlatform vp = su.getViewingPlatform();
    TransformGroup targetTG = vp.getViewPlatformTransform();

    Transform3D t3d = new Transform3D();
    targetTG.getTransform(t3d);
    t3d.setTranslation( new Vector3d(0,1,Z_START));
    targetTG.setTransform(t3d);
  }  // end of positionViewPoint()


  private void addGrabbers()
  /*  The grabbers are created and added to the camera viewpoint as
      a platform geometry. The behavior objects for monitoring the
      keyboard and game pad are also created, since they send 
      operations to the grabbers to move them.
  */
  { 
    // get targetTG, the TG for the viewpoint
    ViewingPlatform vp = su.getViewingPlatform();
    TransformGroup targetTG = vp.getViewPlatformTransform();

    // create the game pad controller
    GamePadController gamePad = new GamePadController();
    // GamePadController gamePad = null;

    // create the grabbers
    Vector3d posnVec = new Vector3d(0, -0.5, -0.9);
    Grabbers grabbers = new Grabbers(posnVec, 0.4f, targetTG, obstacles, gamePad);  
       /* supply grabbers position and each grabber's x-axis offset */

    // add grabbers to the viewpoint
    PlatformGeometry pg = new PlatformGeometry();
    pg.addChild( grabbers.getTG() );
    vp.setPlatformGeometry(pg);


    // set up game pad behavior to catch game pad input
    GamePadBehavior gamePadBeh = new GamePadBehavior(grabbers, gamePad);
    gamePadBeh.setSchedulingBounds(bounds);
    sceneBG.addChild(gamePadBeh);

    // set up keyboard controls to catch keypresses
    KeyBehavior keyBeh = new KeyBehavior(grabbers);
    keyBeh.setSchedulingBounds(bounds);
    sceneBG.addChild(keyBeh);
  } // end of addGrabbers()



  private void addBackground(String fnm)
  /* Create a spherical background. The texture for the 
     sphere comes from fnm. */
  {
    Texture2D tex = loadTexture(BG_DIR + fnm);

    Sphere sphere = new Sphere(1.0f,
			      Sphere.GENERATE_NORMALS_INWARD |
				  Sphere.GENERATE_TEXTURE_COORDS, 8);   // default = 15 (4, 8)
    Appearance backApp = sphere.getAppearance();
	backApp.setTexture( tex );

	BranchGroup backBG = new BranchGroup();
    backBG.addChild(sphere);

    Background bg = new Background();
    bg.setApplicationBounds(bounds);
    bg.setGeometry(backBG);

    sceneBG.addChild(bg);
  }  // end of addBackground()


  private Texture2D loadTexture(String fn)
  // load image from file fn as a texture
  {
    TextureLoader texLoader = new TextureLoader(fn, null);
    Texture2D texture = (Texture2D) texLoader.getTexture();
    if (texture == null)
      System.out.println("Cannot load texture from " + fn);
    else {
      System.out.println("Loaded texture from " + fn);
      texture.setEnable(true);
    }
    return texture;
  }  // end of loadTexture()


} // end of WrapArmsPad3D class